@@ -60,22 +60,22 @@ the clipboard. Some examples:
 .. code:: sh
 
     # Copy an image to the clipboard:
-    kitty +kitten clipboard picture.png
+    kitten clipboard picture.png
 
     # Copy an image and some text to the clipboard:
-    kitty +kitten clipboard picture.jpg text.txt
+    kitten clipboard picture.jpg text.txt
 
     # Copy text from STDIN and an image to the clipboard:
-    echo hello | kitty +kitten clipboard picture.png /dev/stdin
+    echo hello | kitten clipboard picture.png /dev/stdin
 
     # Copy any raster image available on the clipboard to a PNG file:
-    kitty +kitten clipboard -g picture.png
+    kitten clipboard -g picture.png
 
     # Copy an image to a file and text to STDOUT:
-    kitty +kitten clipboard -g picture.png /dev/stdout
+    kitten clipboard -g picture.png /dev/stdout
 
     # List the formats available on the system clipboard
-    kitty +kitten clipboard -g -m . /dev/stdout
+    kitten clipboard -g -m . /dev/stdout
 '''
 
 usage = '[files to copy to/from]'
