@@ -943,10 +943,10 @@ Listen on the specified socket address for control messages. For example,
 UNIX sockets, not associated with a file, like this: :option:`{appname}
 --listen-on`=unix:@mykitty. Environment variables are expanded and relative
 paths are resolved with respect to the temporary directory. To control kitty,
-you can send commands to it with :italic:`{appname} @` using the
-:option:`{appname} @ --to` option to specify this address. Note that if you run
-:italic:`{appname} @` within a kitty window, there is no need to specify the
-:option:`{appname} @ --to` option as it will automatically read from the
+you can send commands to it with :italic:`kitten @` using the
+:option:`kitten @ --to` option to specify this address. Note that if you run
+:italic:`kitten @` within a kitty window, there is no need to specify the
+:option:`kitten @ --to` option as it will automatically read from the
 environment. Note that this will be ignored unless :opt:`allow_remote_control`
 is set to either: :code:`yes`, :code:`socket` or :code:`socket-only`. This can
 also be specified in :file:`kitty.conf`.
