@@ -30,7 +30,6 @@ from .constants import (
     website_url,
 )
 from .fast_data_types import (
-    GLFW_IBEAM_CURSOR,
     GLFW_MOD_ALT,
     GLFW_MOD_SHIFT,
     SingleKey,
@@ -76,7 +75,7 @@ def set_custom_ibeam_cursor() -> None:
     rgba_data2, width2, height2 = load_png_data(data)
     images = (rgba_data, width, height), (rgba_data2, width2, height2)
     try:
-        set_custom_cursor(GLFW_IBEAM_CURSOR, images, 4, 8)
+        set_custom_cursor("beam", images, 4, 8)
     except Exception as e:
         log_error(f'Failed to set custom beam cursor with error: {e}')
 
@@ -270,8 +269,9 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ())
                     wincls, wstate, load_all_shaders, disallow_override_title=bool(args.title))
         boss = Boss(opts, args, cached_values, global_shortcuts)
         boss.start(window_id, startup_sessions)
-        if bad_lines:
-            boss.show_bad_config_lines(bad_lines)
+        if bad_lines or boss.misc_config_errors:
+            boss.show_bad_config_lines(bad_lines, boss.misc_config_errors)
+            boss.misc_config_errors = []
         try:
             boss.child_monitor.main_loop()
         finally:
