@@ -2,6 +2,7 @@
 # License: GPLv3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>
 
 
+import enum
 import re
 import sys
 from functools import lru_cache
@@ -50,6 +51,12 @@ func_with_args = KeyFuncWrapper[FuncArgsType]()
 DELETE_ENV_VAR = '_delete_this_env_var_'
 
 
+class MapType(enum.Enum):
+    MAP = 'map'
+    MOUSE_MAP = 'mouse_map'
+    OPEN_ACTION = 'open_action'
+
+
 class InvalidMods(ValueError):
     pass
 
@@ -280,6 +287,14 @@ def remote_control(func: str, rest: str) -> FuncArgsType:
     return func, args
 
 
+@func_with_args('remote_control_script')
+def remote_control_script(func: str, rest: str) -> FuncArgsType:
+    func, args = shlex_parse(func, rest)
+    if len(args) < 1:
+        log_error('Too few arguments to remote_control_script function')
+    return func, args
+
+
 @func_with_args('nth_os_window', 'nth_window', 'scroll_to_prompt', 'visual_window_select_action_trigger', 'next_layout')
 def single_integer_arg(func: str, rest: str) -> FuncArgsType:
     try:
@@ -388,6 +403,7 @@ def mouse_selection(func: str, rest: str) -> FuncArgsType:
             'word': defines.MOUSE_SELECTION_WORD,
             'line': defines.MOUSE_SELECTION_LINE,
             'line_from_point': defines.MOUSE_SELECTION_LINE_FROM_POINT,
+            'word_and_line_from_point': defines.MOUSE_SELECTION_WORD_AND_LINE_FROM_POINT,
         }
         setattr(mouse_selection, 'code_map', cmap)
     return func, [cmap[rest]]
@@ -863,6 +879,26 @@ def store_multiple(val: str, current_val: Container[str]) -> Iterable[Tuple[str,
         yield val, val
 
 
+def menu_map(val: str, current_val: Container[str]) -> Iterable[Tuple[Tuple[str, ...], str]]:
+    parts = val.split(maxsplit=1)
+    if len(parts) != 2:
+        raise ValueError(f'Ignoring invalid menu action: {val}')
+    if parts[0] != 'global':
+        raise ValueError(f'Unknown menu type: {parts[0]}. Known types: global')
+    start = 0
+    if parts[1].startswith('"'):
+        start = 1
+        idx = parts[1].find('"', 1)
+        if idx == -1:
+            raise ValueError(f'The menu entry name in {val} must end with a double quote')
+    else:
+        idx = parts[1].find(' ')
+        if idx == -1:
+            raise ValueError(f'The menu entry {val} must have an action')
+    location = ('global',) + tuple(parts[1][start:idx].split('::'))
+    yield location, parts[1][idx+1:].lstrip()
+
+
 allowed_shell_integration_values = frozenset({'enabled', 'disabled', 'no-rc', 'no-cursor', 'no-title', 'no-prompt-mark', 'no-complete', 'no-cwd', 'no-sudo'})
 
 
@@ -875,11 +911,11 @@ def shell_integration(x: str) -> FrozenSet[str]:
 
 
 def paste_actions(x: str) -> FrozenSet[str]:
-    s = frozenset({'quote-urls-at-prompt', 'confirm', 'filter'})
+    s = frozenset({'quote-urls-at-prompt', 'confirm', 'filter', 'confirm-if-large', 'replace-dangerous-control-codes', 'replace-newline'})
     q = frozenset(x.lower().split(','))
     if not q.issubset(s):
         log_error(f'Invalid paste actions: {q - s}, ignoring')
-        return q & s or frozenset({'invalid'})
+        return (q & s) or frozenset({'invalid'})
     return q
 
 
@@ -929,7 +965,7 @@ def narrow_symbols(val: str) -> Iterable[Tuple[Tuple[int, int], int]]:
         yield x, int(y or 1)
 
 
-def parse_key_action(action: str, action_type: str = 'map') -> KeyAction:
+def parse_key_action(action: str, action_type: MapType = MapType.MAP) -> KeyAction:
     parts = action.strip().split(maxsplit=1)
     func = parts[0]
     if len(parts) == 1:
@@ -960,7 +996,7 @@ class AliasMap:
         self.aliases.update(aa.aliases)
 
     @lru_cache(maxsize=256)
-    def resolve_aliases(self, definition: str, map_type: str = 'map') -> Tuple[KeyAction, ...]:
+    def resolve_aliases(self, definition: str, map_type: MapType = MapType.MAP) -> Tuple[KeyAction, ...]:
         return tuple(resolve_aliases_and_parse_actions(definition, self.aliases, map_type))
 
 
@@ -976,7 +1012,7 @@ def build_action_aliases(raw: Dict[str, str], first_arg_replacement: str = '') -
 
 
 def resolve_aliases_and_parse_actions(
-    defn: str, aliases: Dict[str, List[ActionAlias]], map_type: str
+    defn: str, aliases: Dict[str, List[ActionAlias]], map_type: MapType
 ) -> Iterator[KeyAction]:
     parts = defn.split(maxsplit=1)
     if len(parts) == 1:
@@ -1016,7 +1052,7 @@ def resolve_aliases_and_parse_actions(
 
 class BaseDefinition:
     no_op_actions = frozenset(('noop', 'no-op', 'no_op'))
-    map_type: str = 'map'
+    map_type: MapType = MapType.MAP
     definition_location: CurrentlyParsing
 
     def __init__(self, definition: str = '') -> None:
@@ -1044,7 +1080,7 @@ def resolve_key_mods(kitty_mod: int, mods: int) -> int:
 
 
 class MouseMapping(BaseDefinition):
-    map_type: str = 'mouse_map'
+    map_type: MapType = MapType.MOUSE_MAP
 
     def __init__(
         self, button: int = 0, mods: int = 0, repeat_count: int = 1, grabbed: bool = False,
