@@ -19,6 +19,7 @@ from typing import (
     Iterator,
     List,
     Optional,
+    Sequence,
     Tuple,
     Union,
     cast,
@@ -104,6 +105,27 @@ def fnmatch_pattern(pat: str) -> 're.Pattern[str]':
     return re.compile(translate(pat))
 
 
+def remote_control_allowed(
+    pcmd: Dict[str, Any], remote_control_passwords: Optional[Dict[str, Sequence[str]]],
+    window: Optional['Window'], extra_data: Dict[str, Any]
+) -> bool:
+    if not remote_control_passwords:
+        return True
+    pw = pcmd.get('password', '')
+    auth_items = remote_control_passwords.get(pw)
+    if pw == '!':
+        auth_items = None
+    if auth_items is None:
+        if '!' in remote_control_passwords:
+            raise PermissionError()
+        return False
+    from .remote_control import password_authorizer
+    pa = password_authorizer(auth_items)
+    if not pa.is_cmd_allowed(pcmd, window, False, extra_data):
+        raise PermissionError()
+    return True
+
+
 class PasswordAuthorizer:
 
     def __init__(self, auth_items: FrozenSet[str]) -> None:
@@ -153,7 +175,7 @@ def is_cmd_allowed(pcmd: Dict[str, Any], window: Optional['Window'], from_socket
         return True
     if 'cancel_async' in pcmd and pcmd.get('async_id'):
         # we allow these without authentication as they are sent on error
-        # conditions and we cant have users prompted for these. The worst side
+        # conditions and we can't have users prompted for these. The worst side
         # effect of a malicious cancel_async request is that it can prevent
         # another async request from getting a result, if it knows the async_id
         # of that request.
@@ -257,14 +279,14 @@ completion=type:file relative:conf kwds:-
 default=rc-pass
 A file from which to read the password. Trailing whitespace is ignored. Relative
 paths are resolved from the kitty configuration directory. Use - to read from STDIN.
-Used if no :option:`kitty @ --password` is supplied. Defaults to checking for the
+Used if no :option:`kitten @ --password` is supplied. Defaults to checking for the
 :file:`rc-pass` file in the kitty configuration directory.
 
 
 --password-env
 default=KITTY_RC_PASSWORD
 The name of an environment variable to read the password from.
-Used if no :option:`kitty @ --password-file` is supplied. Defaults
+Used if no :option:`kitten @ --password-file` is supplied. Defaults
 to checking the environment variable :envvar:`KITTY_RC_PASSWORD`.
 
 
