@@ -64,15 +64,18 @@ from .fast_data_types import (
     get_boss,
     get_click_interval,
     get_options,
+    is_css_pointer_name_valid,
     last_focused_os_window_id,
     mark_os_window_dirty,
     mouse_selection,
     move_cursor_to_mouse_if_in_prompt,
+    pointer_name_to_css_name,
     pt_to_px,
     set_window_logo,
     set_window_padding,
     set_window_render_data,
     update_ime_position_for_window,
+    update_pointer_shape,
     update_window_title,
     update_window_visibility,
     wakeup_main_loop,
@@ -254,11 +257,15 @@ class Watchers:
     on_resize: List[Watcher]
     on_close: List[Watcher]
     on_focus_change: List[Watcher]
+    on_set_user_var: List[Watcher]
+    on_title_change: List[Watcher]
 
     def __init__(self) -> None:
         self.on_resize = []
         self.on_close = []
         self.on_focus_change = []
+        self.on_set_user_var = []
+        self.on_title_change = []
 
     def add(self, others: 'Watchers') -> None:
         def merge(base: List[Watcher], other: List[Watcher]) -> None:
@@ -268,15 +275,20 @@ class Watchers:
         merge(self.on_resize, others.on_resize)
         merge(self.on_close, others.on_close)
         merge(self.on_focus_change, others.on_focus_change)
+        merge(self.on_set_user_var, others.on_set_user_var)
+        merge(self.on_title_change, others.on_title_change)
 
     def clear(self) -> None:
         del self.on_close[:], self.on_resize[:], self.on_focus_change[:]
+        del self.on_set_user_var[:], self.on_title_change[:]
 
     def copy(self) -> 'Watchers':
         ans = Watchers()
         ans.on_close = self.on_close[:]
         ans.on_resize = self.on_resize[:]
         ans.on_focus_change = self.on_focus_change[:]
+        ans.on_set_user_var = self.on_set_user_var[:]
+        ans.on_title_change = self.on_title_change[:]
         return ans
 
     @property
@@ -467,6 +479,18 @@ class GlobalWatchers:
 global_watchers = GlobalWatchers()
 
 
+def replace_control_codes(text: str) -> str:
+    # Replace all control codes other than tab, newline and space with their graphical counterparts
+    def sub(m: re.Match[str]) -> str:
+        c = ord(m.group())
+        if c < 0x20:
+            return chr(0x2400 + c)
+        if c == 0x7f:
+            return '\u2421'
+        return '\u2426'
+    return re.sub(r'[\0-\x08\x0b-\x19\x7f-\x9f]', sub, text)
+
+
 class Window:
 
     window_custom_type: str = ''
@@ -554,21 +578,8 @@ class Window:
     def remote_control_allowed(self, pcmd: Dict[str, Any], extra_data: Dict[str, Any]) -> bool:
         if not self.allow_remote_control:
             return False
-        if not self.remote_control_passwords:
-            return True
-        pw = pcmd.get('password', '')
-        auth_items = self.remote_control_passwords.get(pw)
-        if pw == '!':
-            auth_items = None
-        if auth_items is None:
-            if '!' in self.remote_control_passwords:
-                raise PermissionError()
-            return False
-        from .remote_control import password_authorizer
-        pa = password_authorizer(auth_items)
-        if not pa.is_cmd_allowed(pcmd, self, False, extra_data):
-            raise PermissionError()
-        return True
+        from .remote_control import remote_control_allowed
+        return remote_control_allowed(pcmd, self.remote_control_passwords, self, extra_data)
 
     @property
     def file_transmission_control(self) -> 'FileTransmission':
@@ -586,11 +597,13 @@ class Window:
         self.os_window_id = tab.os_window_id
         self.tabref = weakref.ref(tab)
 
-    def effective_margin(self, edge: EdgeLiteral, is_single_window: bool = False) -> int:
+    def effective_margin(self, edge: EdgeLiteral) -> int:
         q = getattr(self.margin, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
         opts = get_options()
+        tab = self.tabref()
+        is_single_window = tab is not None and tab.has_single_window_visible()
         if is_single_window:
             q = getattr(opts.single_window_margin_width, edge)
             if q > -0.1:
@@ -602,7 +615,14 @@ class Window:
         q = getattr(self.padding, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
-        q = getattr(get_options().window_padding_width, edge)
+        opts = get_options()
+        tab = self.tabref()
+        is_single_window = tab is not None and tab.has_single_window_visible()
+        if is_single_window:
+            q = getattr(opts.single_window_padding_width, edge)
+            if q > -0.1:
+                return pt_to_px(q, self.os_window_id)
+        q = getattr(opts.window_padding_width, edge)
         return pt_to_px(q, self.os_window_id)
 
     def update_effective_padding(self) -> None:
@@ -646,7 +666,7 @@ class Window:
             'pid': self.child.pid,
             'cwd': self.child.current_cwd or self.child.cwd,
             'cmdline': self.child.cmdline,
-            'env': self.child.environ,
+            'env': self.child.environ or self.child.final_env,
             'foreground_processes': self.child.foreground_processes,
             'is_self': is_self,
             'at_prompt': self.at_prompt,
@@ -669,6 +689,7 @@ class Window:
             'env': self.child.environ,
             'cmdline': self.child.cmdline,
             'margin': self.margin.serialize(),
+            'user_vars': self.user_vars,
             'padding': self.padding.serialize(),
         }
         if self.window_custom_type:
@@ -865,8 +886,39 @@ class Window:
         if title:
             title = sanitize_title(title)
         self.override_title = title or None
+        self.call_watchers(self.watchers.on_title_change, {'title': self.title, 'from_child': False})
         self.title_updated()
 
+    @ac(
+        'win', '''
+        Change the title of the active window interactively, by typing in the new title.
+        If you specify an argument to this action then that is used as the title instead of asking for it.
+        Use the empty string ("") to reset the title to default. Use a space (" ") to indicate that the
+        prompt should not be pre-filled. For example::
+
+            # interactive usage
+            map f1 set_window_title
+            # set a specific title
+            map f2 set_window_title some title
+            # reset to default
+            map f3 set_window_title ""
+            # interactive usage without prefilled prompt
+            map f3 set_window_title " "
+        '''
+    )
+    def set_window_title(self, title: Optional[str] = None) -> None:
+        if title is not None and title not in ('" "', "' '"):
+            if title in ('""', "''"):
+                title = ''
+            self.set_title(title)
+            return
+        prefilled = self.title
+        if title in ('" "', "' '"):
+            prefilled = ''
+        get_boss().get_line(
+            _('Enter the new title for this window below. An empty title will cause the default title to be used.'),
+            self.set_title, window=self, initial_value=prefilled)
+
     def set_user_var(self, key: str, val: Optional[Union[str, bytes]]) -> None:
         key = sanitize_control_codes(key).replace('\n', ' ')
         self.user_vars.pop(key, None)  # ensure key will be newest in user_vars even if already present
@@ -876,7 +928,10 @@ class Window:
         if val is not None:
             if isinstance(val, bytes):
                 val = val.decode('utf-8', 'replace')
-            self.user_vars[key] = sanitize_control_codes(val).replace('\n', ' ')
+            self.user_vars[key] = val = sanitize_control_codes(val).replace('\n', ' ')
+            self.call_watchers(self.watchers.on_set_user_var, {'key': key, 'value': val})
+        else:
+            self.call_watchers(self.watchers.on_set_user_var, {'key': key, 'value': None})
 
     # screen callbacks {{{
 
@@ -937,7 +992,7 @@ class Window:
                     'What would you like to do with this URL:\n' + styled(sanitize_url_for_dispay_to_user(url), fg='yellow'),
                     partial(self.hyperlink_open_confirmed, url, cwd),
                     'o:Open', 'c:Copy to clipboard', 'n;red:Nothing', default='o',
-                    window=self,
+                    window=self, title=_('Hyperlink activated'),
                 )
                 return
         get_boss().open_url(url, cwd=cwd)
@@ -1005,6 +1060,7 @@ class Window:
 
     def title_changed(self, new_title: Optional[str], is_base64: bool = False) -> None:
         self.child_title = process_title_from_child(new_title or self.default_title, is_base64)
+        self.call_watchers(self.watchers.on_title_change, {'title': self.child_title, 'from_child': True})
         if self.override_title is None:
             self.title_updated()
 
@@ -1079,9 +1135,14 @@ class Window:
         identifier = sanitize_identifier_pat().sub('', identifier)
         self.screen.send_escape_code_to_child(OSC, f'99;i={identifier};')
 
+
     def set_dynamic_color(self, code: int, value: Union[str, bytes]) -> None:
         if isinstance(value, bytes):
             value = value.decode('utf-8')
+        if code == 22:
+            ret = set_pointer_shape(self.screen, value, self.os_window_id)
+            if ret:
+                self.screen.send_escape_code_to_child(OSC, '22:' + ret)
         color_changes: Dict[DynamicColor, Optional[str]] = {}
         for val in value.split(';'):
             w = DYNAMIC_COLOR_CODES.get(code)
@@ -1133,6 +1194,9 @@ class Window:
     def handle_remote_echo(self, msg: str) -> None:
         from base64 import standard_b64decode
         data = standard_b64decode(msg)
+        # ensure we are not writing any control char back as this can lead to command injection on shell prompts
+        # Any bytes outside the printable ASCII range are removed.
+        data = re.sub(rb'[^ -~]', b'', data)
         self.write_to_child(data)
 
     def handle_remote_ssh(self, msg: str) -> None:
@@ -1186,6 +1250,7 @@ class Window:
                     'A program running in this window wants to clone it into another window.'
                     ' Allow it do so, once?'),
                     partial(self.handle_remote_clone_confirmation, cdata), window=self,
+                    title=_('Allow cloning of window?'),
                 )
             elif ac in ('yes', 'y', 'true'):
                 self.handle_remote_clone_confirmation(cdata, True)
@@ -1245,6 +1310,7 @@ class Window:
             if pop:
                 if self.title_stack:
                     self.child_title = self.title_stack.pop()
+                    self.call_watchers(self.watchers.on_title_change, {'title': self.child_title, 'from_child': True})
                     self.title_updated()
             else:
                 if self.child_title:
@@ -1446,20 +1512,45 @@ class Window:
                 if rest.startswith('//') or scheme in ('mailto', 'irc'):
                     import shlex
                     text = shlex.quote(text)
+        if 'replace-dangerous-control-codes' in opts.paste_actions:
+            text = replace_control_codes(text)
+        if 'replace-newline' in opts.paste_actions:
+            text = text.replace('\n', '\x1bE')
         btext = text.encode('utf-8')
         if 'confirm' in opts.paste_actions:
-            msg = ''
-            limit = 16 * 1024
+            sanitized = replace_control_codes(text)
             if not self.screen.in_bracketed_paste_mode:
-                msg = _('Pasting text into shells that do not support bracketed paste can be dangerous.')
-            elif len(btext) > limit:
+                # \n is converted to \r and \r is interpreted as the enter key
+                # by legacy programs that dont support the full kitty keyboard protocol,
+                # which in the case of shells can lead to command execution.
+                # \eE has the same visual effect as \r\n but without the
+                # command execution risk.
+                sanitized = sanitized.replace('\n', '\x1bE')
+            if sanitized != text:
+                msg = _('The text to be pasted contains terminal control codes.\n\nIf the terminal program you are pasting into does not properly'
+                        ' sanitize pasted text, this can lead to \x1b[31mcode execution vulnerabilities\x1b[39m.\n\nHow would you like to proceed?')
+                get_boss().choose(
+                    msg, partial(self.handle_dangerous_paste_confirmation, btext, sanitized),
+                    's;green:Sanitize and paste', 'p;red:Paste anyway', 'c;yellow:Cancel',
+                    window=self, default='s', title=_('Allow paste?'),
+                )
+                return
+        if 'confirm-if-large' in opts.paste_actions:
+            msg = ''
+            if len(btext) > 16 * 1024:
                 msg = _('Pasting very large amounts of text ({} bytes) can be slow.').format(len(btext))
-            if msg:
-                get_boss().confirm(msg + _(' Are you sure?'), partial(self.handle_paste_confirmation, btext), window=self)
+                get_boss().confirm(msg + _(' Are you sure?'), partial(self.handle_large_paste_confirmation, btext), window=self, title=_(
+                'Allow large paste?'))
                 return
         self.paste_text(btext)
 
-    def handle_paste_confirmation(self, btext: bytes, confirmed: bool) -> None:
+    def handle_dangerous_paste_confirmation(self, unsanitized: bytes, sanitized: str, choice: str) -> None:
+        if choice == 's':
+            self.paste_text(sanitized)
+        elif choice == 'p':
+            self.paste_text(unsanitized)
+
+    def handle_large_paste_confirmation(self, btext: bytes, confirmed: bool) -> None:
         if confirmed:
             self.paste_text(btext)
 
@@ -1723,3 +1814,36 @@ class Window:
         url = docs_url(which)
         get_boss().open_url(url)
     # }}}
+
+
+def set_pointer_shape(screen: Screen, value: str, os_window_id: int = 0) -> str:
+    op, ret = '=', ''
+    if value and value[0] in '><=?':
+        op = value[0]
+        value = value[1:]
+    if op in '=>':
+        for v in value.split(','):
+            if v or op == '=':
+                screen.change_pointer_shape(op, v)
+        if os_window_id and current_focused_os_window_id() == os_window_id:
+            update_pointer_shape(os_window_id)
+    elif op == '<':
+        screen.change_pointer_shape('<', '')
+        if os_window_id and current_focused_os_window_id() == os_window_id:
+            update_pointer_shape(os_window_id)
+    elif op == '?':
+        ans = []
+        for q in value.split(','):
+            if is_css_pointer_name_valid(q):
+                ans.append('1')
+            else:
+                if q == '__default__':
+                    ans.append(pointer_name_to_css_name(get_options().default_pointer_shape))
+                elif q == '__grabbed__':
+                    ans.append(pointer_name_to_css_name(get_options().pointer_shape_when_grabbed))
+                elif q == '__current__':
+                    ans.append(screen.current_pointer_shape())
+                else:
+                    ans.append('0')
+        ret = ','.join(ans)
+    return ret
